--- The utils module
-- @author Oskar Eisemuth
-- @copyright 2017-2020
-- @module commonapi.utils
local utils = {}

local osIsWin = (os.getenv("OS") == "Windows_NT")
local mymodid = nil

local function getOwnModId()
	local mypath = debug.getinfo(2, "S").source:sub(2)
	local mypath = string.gsub(mypath, '\\', '/')
	mymodid = string.match(mypath, "([^/]+)/res/scripts/commonapi2/utils%.lua$")
end

getOwnModId()

--- Creates a log routine with prefix
-- @param prefix prefix for log
-- @return log function
function utils.logPrep(prefix)
	local prefix = prefix..":"
	local logfn = function(...)
		print(prefix, ...)
	end
	return logfn
end

local _log = utils.logPrep("commonapi2.utils")

function utils.logPrepSlow(name)
	local prefix = name.."."
	local logfn = function(...)
		local dinfo = debug.getinfo(2, "n")
		if (dinfo ~= nil) then
			print(prefix..dinfo.name..":",...)
		else
			print(...)
		end
	end
	return logfn
end


--- Split a string by single separator char
-- @param str string to split
-- @param sep char separator
-- @return table
function utils.split(str, sep)
	if string.find(str, sep) == nil then
		return { str }
	end
	local res = {}
	local fpat = "(.-)%" .. sep
	local last_end = 1
	local s, e, cap = str:find(fpat, 1)
	while s do
		if s ~= 1 or cap ~= "" then
			table.insert(res, cap)
		end
		last_end = e+1
		s, e, cap = str:find(fpat, last_end)
	end
	if last_end <= #str then
		cap = str:sub(last_end)
		table.insert(res, cap)
	end
	return res
end



function utils.convertVersion(strversion)
	local version = {}
	local parts = utils.split(strversion, '.')
	for i, s in ipairs(parts) do
		local n = tonumber(s)
		if n ~= nil then
			version[i] = n
		else
			version[i] = 0
		end
	end
	return version
end

function utils.compareVersion(a, b)
	local l = math.max(#a, #b)
	for i = 1, l do
		local ai = a[i] or 0
		local bi = b[i] or 0
		if (ai < bi) then
			return -1
		end
		if (ai > bi) then
			return 1
		end
	end
	return 0
end


--- Is running on windows.
-- @return boolean
function utils.isWindowsOS()
	return osIsWin
end

--- Get a valid tempfile.
-- Returns a valid tempfile name (on windows including temp directory)
-- @return string
function utils.getTempFilename()
	local filename
	if (osIsWin == true) then
		-- Fix crash without Native in TPF2
		--local prefix = os.getenv("TEMP")
		--filename = prefix..os.tmpname()
		filename = os.tmpname()
	else
		filename = os.tmpname()
	end
	return filename
end


--- Returns the data function result of a game file.
-- @param filename A complete filepath to a lua file
-- @param env	Environment table (optional) 
-- @return table
function utils.getFileData(filename, env)
	return utils._getFileDataInternal(filename, env, true)
end

function utils._getFileDataIfExist(filename, env)
	return utils._getFileDataInternal(filename, env, false)
end

function utils._getFileDataInternal(filename, env, errIfMissing)
	local result = nil
	local luafn, err
	local ok

--	print("utils._getFileDataInternal", filename, env)

	if (env == nil) then
		env = {
			data = function()
				return {}
			end
		}
	end

	local debugGetInfo = debug.getinfo
	debug.getinfo = function(...)
		local n = select('#', ...);
		local info = nil
		if (n < 3) then
			local arg1 = select(1, ...);
			local arg2 = select(2, ...);
			if (type(arg1) == "number") then
				arg1 = arg1 + 1
			end

			if (type(arg2) == "string") then
				arg2 = arg2.."S"
			else
				arg2 = "S"
			end
			info = debugGetInfo(arg1, arg2)
		else
			local arg1 = select('1', ...);
			local arg2 = select('2', ...);
			local arg3 = select('3', ...);
			if (type(arg2) == "number") then
				arg2 = arg2 + 2
			end

			if (type(arg3) == "string") then
				arg3 = arg3.."S"
			else
				arg3 = "S"
			end
			info = debugGetInfo(arg1, arg2, arg3)
		end
		if (info ~= nil) then
			if (info["short_src"] == "[C]") then
				return nil
			end
		end
		return info
	end

	setmetatable(env, {__index=_G})

	luafn, err = loadfile(filename, "bt", env)
	if (luafn == nil) then
		if (errIfMissing) then
			_log(err)
		end
		debug.getinfo = debugGetInfo
		return nil
	end

	ok, err = pcall(luafn)
	if (ok == false) then
		_log(err)
		debug.getinfo = debugGetInfo
		return nil
	end

	if (env.data ~= nil) then
		ok, result = pcall(env.data)
		if (ok == false) then
			_log(result)
			debug.getinfo = debugGetInfo
			return nil
		end
	end

	debug.getinfo = debugGetInfo
	return result
end

function utils.quoteParam(filename) 
	return '"'..string.gsub(filename, '"', '\\"')..'"'
end


function utils.fixslashes(filename)
	if (osIsWin == true) then
		return string.gsub(filename, '/', '\\');
	else
		return string.gsub(filename, '\\', '/');
	end
end

local function scanDirNative(nativefs, directory, result)
--	_log("scanDirNative:",  directory)
	local realpath = nativefs.realpath(directory)
	if (realpath ~= nil) then
		realpath = string.gsub(realpath, '\\', '/')..'/'
		for filename, fflags in nativefs.dir(directory) do
			if (fflags == 2) then
				scanDirNative(nativefs, directory..filename, result)
			else
				result[#result+1] = realpath..filename
			end
		end
	end
end

local function scanDirsNative(dirs)
	local result = {}
	local nativefs = commonapi._native.fs
	for id, entry in ipairs( dirs ) do
		scanDirNative(nativefs, entry, result)
	end
	return result
end

local function scanDirsShell(dirs)
	local result = {}
	local p
	
	if (dirs == nil) then
		return result
	end
	
	local ext
	
	if (osIsWin == true) then
		ext = ".cmd"
	else
		ext = ".sh"
	end
	
	local searchpath = string.gsub(package.path, "res/scripts/%?%.lua", "res/scripts/commonapi2/?"..ext)
	
	local exefilename, err = package.searchpath("listdirs", searchpath)
	if (exefilename == nil) then
		_log(err)
		return result
	end
	
	_log(exefilename)
	if (string.match(exefilename, '^%a%a+') ~= nil) then
		_log("Adding local prefix to path")
		exefilename = "./"..exefilename
	end
	
	local tmpfilename = utils.getTempFilename()
	_log(tmpfilename)
	local f = assert(io.open(tmpfilename, "w"))
	for id, entry in ipairs( dirs ) do
		f:write(entry, "\n")
	end
	f:close()
	
	--local pname = '"'..exefilename..'" "'..tmpfilename..'"'
	local pcmd

	if (osIsWin == true) then
		pcmd = '"'..utils.quoteParam(exefilename)..' '..utils.quoteParam(tmpfilename)..'"'
	else
		pcmd = 'sh '..utils.quoteParam(exefilename)..' '..utils.quoteParam(tmpfilename)
	end
	
	_log("Trying to run "..pcmd)
	p = assert(io.popen(pcmd))
	for line in p:lines() do
		result[#result+1] = string.gsub(line, '\\', '/')
	end
	p:close()
	return result
end


--- Scan directories.
-- @param dirs A complete list of filepaths
-- @treturn {string,...} result of directory scan
function utils.scanDirs(dirs)
	local result = {}
	local p
	
	
	local testtimer = os.clock()
	if (dirs == nil) then
		return result
	end
	if (commonapi._native ~= nil) then
		result = scanDirsNative(dirs)
	else
		result = scanDirsShell(dirs)
	end

	_log(string.format("elapsed time for scanDirs: %.2f  (files %d)", os.clock() - testtimer, #result))
	return result
end

--- Create a ordered table
-- @return {table} a table with ordered key value iteration
function utils.createOrderedTable()
	local t = {}
	local mt = {
		__index = {
			_keyorder = {},
			_delete = function(self, key)
				if self[key] == nil then
					return
				end
				self[key] = nil
				local keyorder = self._keyorder
				for i, k in ipairs(keyorder) do
					if k == key then
						table.remove(keyorder, i)
					end
				end
			end,
			_getKeyIndex = function(self, key)
				local keyorder = self._keyorder
				for i, k in ipairs(keyorder) do
					if k == key then
						return i
					end
				end
				return nil
			end,
		},
		__newindex = function(self, key, value)
			rawset(self, key, value)
			if (value == nil) then
				return
			end
			local keyorder = self._keyorder
			keyorder[#keyorder+1] = key
		end,
		__pairs = function(tbl)
			local i = 0
			local function pairs_iter(self)
				i = i + 1
				local key = self._keyorder[i]
				if (key) then
					return key, self[key]
				end
				return
			end
			return pairs_iter, tbl, nil
		end
	}
	setmetatable(t, mt)
	return t
end

local function deepcopy(o)
	if ( type(o) == "table" ) then
		local newtable = {}
		for k, v in pairs(o) do
			if ( type(v) == "table" ) then
				newtable[k] = utils.deepcopy(v)
			else
				newtable[k] = v
			end
		end
		return newtable
	end
	return o
end

--- Copy a table
-- @return {table} a table deep copy
function utils.deepcopy(o)
	return deepcopy(o)
end



local function tryMetatablePairsAtUserdata(mt)
	if (mt.pairs ~= nil) then
		return true
	end
	return false
end

local function tryUserdataObjGet(o, k)
	return o[k]
end

--- Copy a userdata/table to table
-- @return {table} a table deep copy
function utils.deepcopy2table(o)
	if ( type(o) == "table" ) then
		local newtable = {}
		for k, v in pairs(o) do
			if ( type(v) == "table" or type(v) == "userdata") then
				newtable[k] = utils.deepcopy2table(v)
			else
				newtable[k] = v
			end
		end
		return newtable
	elseif (type(o) == "userdata" ) then
		local mt = getmetatable(o)
		if (mt == nil) then
			return tostring(o)
		end

		local newtable = {}
		local statuspairsallowed, pairsallowed = pcall(tryMetatablePairsAtUserdata, mt)
		if (statuspairsallowed and mt.pairs ~= nil) then
			for k, v in pairs(o) do
				newtable[k] = utils.deepcopy2table(v)
			end
		elseif(mt.__members ~= nil) then
			local keys = mt.__members
			for i = 1, #keys do
				local k = keys[i]
				local ok, v = pcall(tryUserdataObjGet, o, k)
				if (ok) then
					newtable[k] = utils.deepcopy2table(v)
				end
			end
		end
		return newtable
	end
	return o
end


function utils.forceStrict()
	local mt = getmetatable(_G)
	if (mt == nil) then
		mt = {}
		setmetatable(_G, mt)
	end

	function mt.__newindex(t, key, value)
		local info = debug.getinfo(2, "S")
		if (info ~= nil) then
			if info.what == "main" or  info.what == "C" then
				rawset(t, key, value)
				return
			end
		end
		print("")
		print("STRICT: Attempt to write to global undeclared variable "..key)
		print(debug.traceback(nil, 2))
		rawset(t, key, value)
	end
	
	function mt.__index(_, key)
		print("")
		print("STRICT: Attempt to read global undeclared variable "..key)
		print(debug.traceback(nil, 2))
	end
end

function utils._getFunctionLocal(fnc, varname)
	local i = 1
	if fnc == nil then
		return nil
	end
	repeat
		local name, val = debug.getupvalue (fnc, i)
		if name == varname then
			return val
		end
		i = i + 1
	until not name
	return nil
end


function utils._setFunctionLocal(fnc, varname, newvalue)
	local ok = false
	if fnc == nil then
		return nil
	end
	local info = debug.getinfo(fnc, 'u')
	local upvalues = info.nups
	for i = 1, upvalues do
		local name, val = debug.getupvalue(fnc, i)
		if name == varname then
			if (debug.setupvalue(fnc, i, newvalue)) then
				ok = true
			end
		end
	end
	return ok
end

function utils._dumpFunctionLocal(fnc, varname)
	local i = 1
	repeat
		local name, val = debug.getupvalue (fnc, i)
		print(name, val)
		i = i + 1
	until not name
	return nil
end





local detectedLanguage = nil
local _i18nLocal = {}
local _translateIdMap = {}

function utils._detectLanguage()
	local lang = string.match(debug.getinfo(_, "S").source, '%"([%w_]+)%", s%) end$')
	if (lang ~= nil) then
		if (detectedLanguage == lang) then
			_log("Detected language:", lang, "(same)")
		else
			_log("Detected language:", lang, "(new)")
			detectedLanguage = lang;
		end
	else
		_log("Detected language:", detectedLanguage, "(none=keepold)")
	end
end

function utils._addTranslateIdAlias(idalias, id)
	_translateIdMap[idalias] = id
end

-- Translate a string of modid
function utils.translateString(modid, s)
	local language

	if (_locale ~= nil and _locale:len() > 0) then
		language = _locale
	elseif (commonapi._native ~= nil) then
		language = commonapi._native.getLanguage()
	else
		if (_locale ~= nil and _locale:len() > 0) then
			language = _locale
		else
			if (detectedLanguage == nil) then
				language = string.match(debug.getinfo(_, "S").source, '%"([%w_]+)%", s%) end$') or "en"
			else
				language = detectedLanguage
			end
		end
	end

	if (_i18nLocal[modid] ~= nil) then
		if (translate ~= nil) then
			return translate(modid, language, s)
		elseif (translateModStr ~= nil) then
			return translateModStr(modid, language, s)
		end
	else
		if (_translateIdMap[modid] ~= nil) then
			if (translate ~= nil) then
				return translate(_translateIdMap[modid], language, s)
			elseif (translateModStr ~= nil) then
				return translateModStr(_translateIdMap[modid], language, s)
			end
		end
	end
	
--	print("utils.translateString", language, modid, s, "=", translate2(modid, language, s))
	if getTextRes then
		return getTextRes(s)
	end
	return s
-- return translate2(modid, language, s)
end

function utils._t(s)
	local language
	if (commonapi._native ~= nil) then
		language = commonapi._native.getLanguage()
	else
		if (_locale ~= nil and _locale:len() > 0) then
			language = _locale
		else
			local getinfo = debug.getinfo(_, "S")
			if (getinfo and getinfo.source) then
				language = string.match(debug.getinfo(_, "S").source, '%"([%w_]+)%", s%) end$') or "en"
			else
				language = "en"
			end
		end
	end

--	return translate2(mymodid, language, s)
-- end
	if (translate ~= nil) then
		local txt = translate(mymodid, language, s)
		if txt ~= s then
			return txt
		end
	elseif (translateModStr ~= nil) then
		local txt = translateModStr(mymodid, language, s)
		if txt ~= s then
			return txt
		end
	end
	if getTextRes then
		return getTextRes(s)
	end
	return s
end

local function prepareLanguageFunctions()
	if (setStrings == nil) then
		require "mod"
	end
	if (type(setStrings) == "function") then
		_i18nLocal = utils._getFunctionLocal(setStrings, "i18n")
	else
		_log("ERROR: Can't get i18n local from setStrings, setStrings is not a function !!!")
	end
	if (_i18nLocal == nil) then
		_log("ERROR: Can't get i18n local from setStrings!!!")
		_i18nLocal = {}
	end
end


function utils.getDirLocalMods()
	local _native = commonapi._native
	local userpath = _native.paths.getUserPath()
	if (userpath == nil) then
		_log("No UserPath found?")
		return
	end
	userpath = string.gsub(userpath, '\\', '/')
	if (userpath:sub(-1) == '/') then
		userpath = userpath:sub(1, -2)
	end
	local path = userpath .. '/mods'
	_native.fs.mkdir(path)
	return path
end

function utils.getDirGameMods()
	local _native = commonapi._native
	local gamepath = _native.paths.getGamePath()
	if (gamepath == nil) then
		_log("No GamePath found?")
		return
	end
	gamepath = string.gsub(gamepath, '\\', '/')
	if (gamepath:sub(-1) == '/') then
		gamepath = gamepath:sub(1, -2)
	end
	local path = gamepath .. '/mods'
	_native.fs.mkdir(path)
	return path
end






local function serializeSettingsData(o, prefix)
	local typ = type(o)
	if typ == "nil" then
		return "nil"
	elseif typ == "boolean" then
		return tostring(o)
	elseif typ == "number" then
		return tostring(o)
	elseif typ == "string" then
		return string.format("%q", o)
	elseif typ == "table" then
		local newprefix = prefix.."\t"
		local seenindex = {}
		local ret = {}

		ret[#ret+1] = "{\n"
		ret[#ret+1] = newprefix
		for k, v in ipairs(o) do
			seenindex[k] = true
			ret[#ret+1] = serializeSettingsData(v, newprefix)
			ret[#ret+1] = ", "
		end

		if (#seenindex > 0) then
			ret[#ret+1] = "\n"
			ret[#ret+1] = newprefix
		end

		for k, v in pairs(o) do
			if (seenindex[k] == nil) then
				local ktype = type(k)
				if (ktype == "string" and string.find(k, "^[_%a][_%w]*$")) then
					ret[#ret+1] = k
					ret[#ret+1] = " = "
				else
					ret[#ret+1] = "["
					ret[#ret+1] = serializeSettingsData(k, newprefix)
					ret[#ret+1] = "] = "
				end
				ret[#ret+1] = serializeSettingsData(v, newprefix)
				ret[#ret+1] = ",\n"
				ret[#ret+1] = newprefix
			end
		end
		ret[#ret+1] = "\n"
		ret[#ret+1] = prefix
		ret[#ret+1] = "}"
		return table.concat(ret)
	elseif typ == "function" then
		return "nil"
	else
		error("serializeSettingsData can't serialize " .. typ)
	end
end

function utils.dmpglobal(s)
	local options = s or {}
	local function processhide(item, path)
		if path[#path] == 'commonapi' then return "<commonapi hidden>" end
		--if path[#path] == 'game' then return "<game hidden>" end
		if path[#path] == 'package' then return "<package hidden>" end
		return item
	end
	options.process = processhide
	commonapi.dmp(_G, options)
end

function utils.serializeData(o)
	local serData = "return "..serializeSettingsData(o, "").."\n"
	return serData
end

function utils.serializeDataToFn(o)
	local serData = "function data()\nreturn "..serializeSettingsData(o, "").."\nend\n"
	return serData
end

prepareLanguageFunctions()

return utils
